import logging

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

# The background is set with 40 plus the number of the color, and the foreground with 30

# These are the sequences need to get colored ouput
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"


def formatter_message(message, use_color=True):
    if use_color:
        message = message.replace("$RESET", RESET_SEQ).replace("$BOLD", BOLD_SEQ)
    else:
        message = message.replace("$RESET", "").replace("$BOLD", "")
    return message


COLORS = {
    'WARNING': YELLOW,
    'INFO': WHITE,
    'DEBUG': BLUE,
    'CRITICAL': YELLOW,
    'ERROR': RED
}


class ColoredFormatter(logging.Formatter):
    def __init__(self, msg, use_color=True):
        logging.Formatter.__init__(self, msg)
        self.use_color = use_color

    def format(self, record):
        levelname = record.levelname
        if self.use_color and levelname in COLORS:
            levelname_color = COLOR_SEQ % (30 + COLORS[levelname]) + levelname + RESET_SEQ
            record.levelname = levelname_color
        return logging.Formatter.format(self, record)


logger = logging.getLogger()
ch = logging.StreamHandler()
formatter = ColoredFormatter('[%(module)s.%(funcName)s] (L.%(lineno)d) - %(levelname)s - %(asctime)s - %(message)s')
logger.propagate = False
# logger.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logger.addHandler(ch)

ch = logging.StreamHandler()
ch.setFormatter(formatter)
event_logger = logging.getLogger('event_logger')
event_logger.setLevel(logging.ERROR)
event_logger.propagate = False
event_logger.addHandler(ch)



from typing import List
import requests

from .api_resource import ApiResource
from . import client
from .utils import PrevisionException, EventTuple, zip_to_pandas


class ValidationPrediction(ApiResource):
    """ Prediction object represent usecase prediction bulk resource that will be explored by Prevision.io platform.

    """

    resource = 'validation-prediction'

    def __init__(self, _id: str,usecase_id: str, usecase_version_id: str, project_id: str,  state='running' , model_id=None, model_name=None,
                 dataset_id=None, download_available=False, score=None, duration=None, predictions_count=None, **kwargs):

        self._id = _id
        self.usecase_id = usecase_id
        self.usecase_version_id = usecase_version_id
        self.project_id = project_id
        self.model_id = model_id
        self.model_name = model_name
        self.dataset_id = dataset_id
        self.download_available = download_available
        self.score = score
        self.duration = duration
        self.predictions_count = predictions_count
        self._state = state

    @classmethod
    def from_id(cls, _id: str) -> 'ValidationPrediction':
        """Get a prediction from the platform by its unique id.

        Args:
            _id (str): Unique id of the usecase prediction bulk to retrieve

        Returns:
            :class:`.Prediction`: Fetched prediction

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        return cls(**super()._from_id(specific_url='/{}/{}'.format(cls.resource, _id)))

    @property
    def status(self):
        status = self._status
        return status['state']

    def delete(self):
        """Delete a prediction from the platform by its unique id.

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        response = client.request(endpoint='/{}/{}'.format(self.resource, self._id),
                                  method=requests.delete,
                                  message_prefix='Prediction deletion')
        return response

    def get_result(self):
        """Delete a prediction from the platform by its unique id.

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        self._wait_for_prediction()
        url = '/{}/{}/download'.format(self.resource, self._id)
        pred_response = client.request(url,
                                       method=requests.get,
                                       message_prefix='Predictions download')

        return zip_to_pandas(pred_response)
        return pred_response

    def _wait_for_prediction(self):
        """ Wait for a specific prediction to finish.

        Args:
            predict_id (str): Unique id of the prediction to wait for
        """
        specific_url = '/{}/{}'.format(self.resource, self._id)
        client.event_manager.wait_for_event(self._id,
                                            specific_url,
                                            EventTuple('PREDICTION_UPDATE', 'state', 'done', [('state', 'failed')]),
                                            specific_url=specific_url)


class DeploymentPrediction(ApiResource):
    """ Prediction object represent usecase prediction bulk resource that will be explored by Prevision.io platform.

    """

    resource = 'deployment-predictions'

    def __init__(self, _id: str, project_id: str, deployment_id: str, state='running', main_model_id=None,
                 challenger_model_id=None, **kwargs):

        self._id = _id
        self.project_id = project_id
        self.deployment_id = deployment_id
        self.main_model_id = main_model_id
        self.challenger_model_id = challenger_model_id
        self._state = state
        for k, v in kwargs.items():
            #print("k============", k)
            #print("v============", v)
            self.__setattr__(k, v)

    @classmethod
    def from_id(cls, _id: str) -> 'DeploymentPrediction':
        """Get a prediction from the platform by its unique id.

        Args:
            _id (str): Unique id of the usecase prediction bulk to retrieve

        Returns:
            :class:`.DeploymentPrediction`: Fetched prediction

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing result
        """
        return cls(**super()._from_id(specific_url='/{}/{}'.format(cls.resource, _id)))

    def get_result(self):
        specific_url = '/{}/{}'.format(self.resource, self._id)
        client.event_manager.wait_for_event(self._id,
                                            specific_url,
                                            EventTuple('DEPLOYMENT_PREDICTION_UPDATE', 'main_model_prediction_state', 'done',
                                                       [('main_model_prediction_state', 'failed')]),
                                            specific_url=specific_url)
        #
        url = '/{}/{}/download'.format(self.resource, self._id)
        pred_response = client.request(url,
                                       method=requests.get,
                                       message_prefix='Predictions download')

        return zip_to_pandas(pred_response)

    def get_challenger_result(self):
        if self.challenger_model_id is None:
            PrevisionException('Challenger data not availbale for this prediction')
        specific_url = '/{}/{}'.format(self.resource, self._id)
        client.event_manager.wait_for_event(self._id,
                                            specific_url,
                                            EventTuple('DEPLOYMENT_PREDICTION_UPDATE', 'challenger_model_prediction_state', 'done',
                                                       [('challenger_model_prediction_state', 'failed')]),
                                            specific_url=specific_url)
        #
        url = '/{}/{}/download'.format(self.resource, self._id)
        pred_response = client.request(url,
                                       method=requests.get,
                                       message_prefix='Predictions download')

        return zip_to_pandas(pred_response)



from enum import Enum
from typing import Dict, List, Union
from previsionio.dataset import Dataset
import requests
from .usecase_config import DataType, UsecaseConfig, TypeProblem, YesOrNo, YesOrNoOrAuto
from .prevision_client import client
from .utils import parse_json, EventTuple, to_json
from .model import TextSimilarityModel
from .usecase_version import BaseUsecaseVersion
import previsionio as pio


class ModelEmbedding(Enum):
    """Embedding models for Text Similarity"""
    TFIDF = 'tf_idf'
    """Term Frequency - Inverse Document Frequency"""
    Transformer = 'transformer'
    """Transformer"""
    TransformerFineTuned = 'transformer_fine_tuned'
    """fine tuned Transformer"""


class TextSimilarityModels(Enum):
    """Similarity search models for Text Similarity"""
    BruteForce = 'brute_force'
    """Brute force search"""
    ClusterPruning = 'cluster_pruning'
    """Cluster Pruning"""
    IVFOPQ = 'ivfopq'
    """InVerted File system and Optimized Product Quantization"""
    HKM = 'hkm'
    """Hierarchical K-Means"""
    LSH = 'lsh'
    """Locality Sensitive Hashing"""


class TextSimilarityLang(Enum):
    Auto = 'auto'
    French = 'fr'
    English = 'en'


class Preprocessing(object):
    config = {}

    def __init__(self,
                 word_stemming: YesOrNo = YesOrNo.Yes,
                 ignore_stop_word: YesOrNoOrAuto = YesOrNoOrAuto.Auto,
                 ignore_punctuation: YesOrNo = YesOrNo.Yes):
        self.word_stemming = word_stemming
        self.ignore_stop_word = ignore_stop_word
        self.ignore_punctuation = ignore_punctuation


class ModelsParameters(UsecaseConfig):
    """ Training configuration that holds the relevant data for a usecase description:
    the wanted feature engineering, the selected models, the training speed...

    Args:
        preprocessing (Preprocessing, optional): Dictionary of the text preprocessings to be applied
            (only for "tf_idf" embedding model),

            - *word_stemming*: default to "yes"
            - *ignore_stop_word*: default to "auto", choice will be made depending on if the
              text descriptions contain full sentences or not
            - *ignore_punctuation*: default to "no".
        model_embedding (ModelEmbedding, optional): Name of the embedding model to be used
            (among: "tf_idf", "transformer", "transformer_fine_tuned").
        models (list(TextSimilarityModels), optional): Names of the searching models to be used (among:
            "brute_force", "cluster_pruning", "ivfopq", "hkm", "lsh").
    """

    config = {
        'model_embedding': 'model_embedding',
        'preprocessing': 'preprocessing',
        'models': 'models'
    }

    def __init__(self,
                 model_embedding: ModelEmbedding = ModelEmbedding.TFIDF,
                 preprocessing: Preprocessing = Preprocessing(),
                 models: List[TextSimilarityModels] = [TextSimilarityModels.BruteForce]):
        self.model_embedding = model_embedding
        if isinstance(preprocessing, Preprocessing):
            self.preprocessing = preprocessing
        elif preprocessing == {}:
            self.preprocessing = preprocessing
        else:
            self.preprocessing = Preprocessing(**preprocessing)
        self.models = models


class ListModelsParameters(UsecaseConfig):

    config = {
        'models_parameters': 'models_params'
    }

    def __init__(self, models_parameters: Union[List[ModelsParameters], List[Dict], None] = None):

        if models_parameters is None:

            models_parameters_1 = ModelsParameters(ModelEmbedding.TFIDF,
                                                   Preprocessing(),
                                                   [TextSimilarityModels.BruteForce,
                                                    TextSimilarityModels.ClusterPruning])
            models_parameters_2 = ModelsParameters(ModelEmbedding.Transformer,
                                                   Preprocessing(),
                                                   [TextSimilarityModels.BruteForce, TextSimilarityModels.IVFOPQ])
            models_parameters_3 = ModelsParameters(ModelEmbedding.TransformerFineTuned,
                                                   Preprocessing(),
                                                   [TextSimilarityModels.BruteForce, TextSimilarityModels.IVFOPQ])
            models_parameters = [models_parameters_1, models_parameters_2, models_parameters_3]
        self.models_parameters = []
        for element in models_parameters:
            if isinstance(element, ModelsParameters):
                self.models_parameters.append(element)
            else:
                self.models_parameters.append(ModelsParameters(**element))


class DescriptionsColumnConfig(UsecaseConfig):
    """ Description Column configuration for starting a usecase: this object defines
    the role of specific columns in the dataset.

    Args:
        content_column (str, required): Name of the column containing the text descriptions in the
            description dataset.
        id_column (str, optional): Name of the id column in the description dataset.
    """

    config = {
    }

    def __init__(self,
                 content_column,
                 id_column):
        self.content_column = content_column
        self.id_column = id_column


class QueriesColumnConfig(UsecaseConfig):
    """ Description Column configuration for starting a usecase: this object defines
    the role of specific columns in the dataset.

    Args:
        content_column (str, required): Name of the column containing the text queries in the
            description dataset.
        id_column (str, optional): Name of the id column in the description dataset.
    """

    # config = {
    #     'queries_dataset_content_column': 'queriesDatasetContentColumn',
    #     'queries_dataset_matching_id_description_column': 'queriesDatasetMatchingIdDescriptionColumn',
    #     'queries_dataset_id_column': 'queriesDatasetIdColumn',
    # }

    def __init__(self,
                 queries_dataset_content_column,
                 queries_dataset_matching_id_description_column,
                 queries_dataset_id_column=None):
        self.queries_dataset_content_column = queries_dataset_content_column
        self.queries_dataset_matching_id_description_column = queries_dataset_matching_id_description_column
        self.queries_dataset_id_column = queries_dataset_id_column


class TextSimilarity(BaseUsecaseVersion):
    """ A text similarity usecase version """

    default_metric: pio.metrics.TextSimilarity = pio.metrics.TextSimilarity.accuracy_at_k
    default_top_k: int = 10
    data_type: DataType = DataType.Tabular
    training_type: TypeProblem = TypeProblem.TextSimilarity
    resource: str = 'usecase-versions'
    model_class = TextSimilarityModel

    def __init__(self, **usecase_info):
        super().__init__(**usecase_info)
        usecase_version_params = usecase_info['usecase_version_params']
        self.metric: pio.metrics.TextSimilarity = pio.metrics.TextSimilarity(
            usecase_version_params.get('metric', self.default_metric))
        self.top_k: int = usecase_version_params.get('top_K', self.default_top_k)
        self.lang: TextSimilarityLang = TextSimilarityLang(usecase_version_params.get('lang', TextSimilarityLang.Auto))

        self.description_column_config = DescriptionsColumnConfig(
            content_column=usecase_version_params.get('content_column'),
            id_column=usecase_version_params.get('id_column'))
        if usecase_info.get('queries_dataset_id'):
            self.queries_dataset = usecase_info.get('queries_dataset_id')
            content_column = usecase_version_params.get('queries_dataset_content_column')
            matching_id = usecase_version_params.get('queries_dataset_matching_id_description_column')
            queries_dataset_id_column = usecase_version_params.get('queries_dataset_id_column', None)
            self.queries_column_config = QueriesColumnConfig(queries_dataset_content_column=content_column,
                                                             queries_dataset_matching_id_description_column=matching_id,
                                                             queries_dataset_id_column=queries_dataset_id_column)
        else:
            self.queries_dataset = None
            self.queries_column_config = None
        models_parameters = usecase_version_params.get('models_params')
        self.models_parameters = ListModelsParameters(models_parameters=models_parameters)

        self._id: str = usecase_info['_id']
        self.usecase_id: str = usecase_info['usecase_id']
        self.project_id: str = usecase_info['project_id']
        self.version: int = usecase_info.get('version', 1)
        self._usecase_info = usecase_info
        self.dataset_id: str = usecase_info['dataset_id']
        self.predictions = {}
        self.predict_token = None

        self._models = {}

    @classmethod
    def from_id(cls, _id: str) -> 'TextSimilarity':
        return cls(**super()._from_id(_id))

    @classmethod
    def load(cls, pio_file: str) -> 'TextSimilarity':
        return cls(**super()._load(pio_file))

    @classmethod
    def _fit(
        cls,
        project_id: str,
        name: str,
        dataset: Dataset,
        description_column_config: DescriptionsColumnConfig,
        metric: pio.metrics.TextSimilarity = pio.metrics.TextSimilarity.accuracy_at_k,
        top_k: int = 10,
        lang: TextSimilarityLang = TextSimilarityLang.Auto,
        queries_dataset: Dataset = None,
        queries_column_config: QueriesColumnConfig = None,
        models_parameters: ListModelsParameters = ListModelsParameters(),
        **kwargs
    ) -> 'TextSimilarity':
        """ Start a supervised usecase training with a specific training configuration
        (on the platform).

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`): Reference to the dataset
                object to use for as training dataset
            description_column_config (:class:`.DescriptionsColumnConfig`): Description column configuration
                (see the documentation of the :class:`.DescriptionsColumnConfig` resource for more details
                on each possible column types)
            metric (str, optional): Specific metric to use for the usecase (default: ``None``)
            queries_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a queries dataset (default: ``None``)

        Returns:
            :class:`.TextSimilarity`: Newly created supervised usecase object
        """

        training_args = to_json(description_column_config)
        assert isinstance(training_args, Dict)
        if queries_column_config:
            training_args.update(to_json(queries_column_config))
        training_args.update(to_json(models_parameters))

        if queries_dataset:
            if isinstance(queries_dataset, str):
                training_args['queries_dataset_id'] = queries_dataset
            else:
                training_args['queries_dataset_id'] = queries_dataset.id

        if not metric:
            metric = cls.default_metric
        if not top_k:
            top_k = cls.default_top_k
        training_args['metric'] = metric.value
        training_args['top_k'] = top_k
        training_args['lang'] = lang.value
        if isinstance(dataset, str):
            dataset_id = dataset
        else:
            dataset_id = dataset.id

        data = dict(name=name, dataset_id=dataset_id, **training_args)

        endpoint = '/projects/{}/{}/{}/{}'.format(project_id, 'usecases', cls.data_type.value, cls.training_type.value)
        start = client.request(endpoint,
                               method=requests.post,
                               data=data,
                               content_type='application/json',
                               message_prefix='Text similarity usecase start')

        start_response = parse_json(start)
        usecase = cls.from_id(start_response['_id'])
        events_url = '/{}/{}'.format(cls.resource, start_response['_id'])
        assert pio.client.event_manager is not None
        pio.client.event_manager.wait_for_event(usecase._id,
                                                cls.resource,
                                                EventTuple('USECASE_VERSION_UPDATE', 'state', 'running',
                                                           [('state', 'failed')]),
                                                specific_url=events_url)
        return usecase

    def new_version(
        self,
        description: str = None,
        dataset: Dataset = None,
        description_column_config: DescriptionsColumnConfig = None,
        metric: pio.metrics.TextSimilarity = None,
        top_k: int = None,
        lang: TextSimilarityLang = TextSimilarityLang.Auto,
        queries_dataset: Dataset = None,
        queries_column_config: Union[QueriesColumnConfig, None] = None,
        models_parameters: ListModelsParameters = None,
        **kwargs
    ) -> 'TextSimilarity':
        """ Start a text similarity usecase training to create a new version of the usecase (on the
        platform): the training configs are copied from the current version and then overridden
        for the given parameters.

        Args:
            description (str, optional): additional description of the version
            dataset (:class:`.Dataset`, :class:`.DatasetImages`, optional): Reference to the dataset
                object to use for as training dataset
            description_column_config (:class:`.DescriptionsColumnConfig`, optional): Column configuration for the
                usecase (see the documentation of the :class:`.ColumnConfig` resource for more details on each possible
                column types)
            metric (metrics.TextSimilarity, optional): Specific metric to use for the usecase (default: ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`, optional): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)
        Returns:
            :class:`.TextSimilarity`: Newly created text similarity usecase version object (new version)
        """

        if not dataset:
            dataset_id = self.dataset_id
        else:
            dataset_id = dataset.id

        if not description_column_config:
            description_column_config = self.description_column_config

        if not metric:
            metric = self.metric
        if not top_k:
            top_k = self.top_k
        if not lang:
            lang = self.lang
        if not queries_dataset:
            queries_dataset_id = self.queries_dataset
        else:
            queries_dataset_id = queries_dataset.id

        if not queries_column_config:
            queries_column_config = self.queries_column_config

        if not models_parameters:
            models_parameters = self.models_parameters

        training_args = to_json(description_column_config)
        assert isinstance(training_args, Dict)
        if queries_column_config:
            training_args.update(to_json(queries_column_config))
        training_args.update(to_json(models_parameters))

        if queries_dataset_id:
            training_args['queries_dataset_id'] = queries_dataset_id

        training_args['metric'] = metric.value
        training_args['top_k'] = top_k
        training_args['lang'] = lang.value

        data = dict(dataset_id=dataset_id, **training_args)

        if description:
            data["description"] = description

        endpoint = "/usecases/{}/versions".format(self.usecase_id)
        resp = client.request(endpoint=endpoint,
                              data=data,
                              method=requests.post,
                              content_type='application/json',
                              message_prefix='Text similarity usecase start')

        start_response = parse_json(resp)
        usecase = self.from_id(start_response['_id'])
        events_url = '/{}/{}'.format(self.resource, start_response['_id'])
        assert pio.client.event_manager is not None
        pio.client.event_manager.wait_for_event(usecase._id,
                                                self.resource,
                                                EventTuple('USECASE_VERSION_UPDATE', 'state', 'running',
                                                           [('state', 'failed')]),
                                                specific_url=events_url)
        return usecase



# -*- coding: utf-8 -*-
from __future__ import print_function
from typing import Dict, Tuple, Union

import requests
from previsionio.usecase_config import ColumnConfig, DataType, TypeProblem
from previsionio.dataset import Dataset, DatasetImages
from . import TrainingConfig
from . import metrics
from .usecase_version import ClassicUsecaseVersion
from .model import RegressionModel, ClassificationModel, MultiClassificationModel
from .utils import EventTuple, parse_json, to_json
from .prevision_client import client
import previsionio as pio

MODEL_CLASS_DICT = {
    TypeProblem.Regression: RegressionModel,
    TypeProblem.Classification: ClassificationModel,
    TypeProblem.MultiClassification: MultiClassificationModel
}


class Supervised(ClassicUsecaseVersion):

    """ A supervised usecase version, for tabular data """

    data_type = DataType.Tabular

    def __init__(self, **usecase_info):
        super().__init__(**usecase_info)
        self.holdout_dataset_id: Union[str, None] = usecase_info.get('holdout_dataset_id', None)

        self.model_class = MODEL_CLASS_DICT.get(self.training_type, RegressionModel)

    @classmethod
    def from_id(cls, _id: str) -> 'Supervised':
        """Get a supervised usecase from the platform by its unique id.

        Args:
            _id (str): Unique id of the usecase to retrieve
            version (int, optional): Specific version of the usecase to retrieve
                (default: 1)

        Returns:
            :class:`.Supervised`: Fetched usecase

        Raises:
            PrevisionException: Invalid problem type or any error while fetching
                data from the platform or parsing result
        """
        return cls(**super()._from_id(_id))

    @classmethod
    def load(cls, pio_file: str) -> 'Supervised':
        return cls(**super()._load(pio_file))

    @classmethod
    def _fit(cls, project_id: str, name: str, data_type: DataType, training_type: TypeProblem,
             dataset: Union[Dataset, Tuple[Dataset, DatasetImages]], column_config: ColumnConfig,
             metric: metrics.Enum, holdout_dataset: Dataset = None,
             training_config: TrainingConfig = TrainingConfig(), **kwargs) -> 'Supervised':
        """ Start a supervised usecase training with a specific training configuration
        (on the platform).

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`, :class:`.DatasetImages`): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`): Column configuration for the usecase
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (str, optional): Specific metric to use for the usecase (default: ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)

        Returns:
            :class:`.Supervised`: Newly created supervised usecase object
        """
        training_args = to_json(training_config)
        assert isinstance(training_args, Dict)
        training_args.update(to_json(column_config))

        if holdout_dataset:
            if isinstance(holdout_dataset, str):
                training_args['holdout_dataset_id'] = holdout_dataset
            else:
                training_args['holdout_dataset_id'] = holdout_dataset.id

        assert metric

        if isinstance(dataset, str):
            dataset_id = dataset
        elif isinstance(dataset, tuple):
            dataset_id = [d.id for d in dataset]
        else:
            dataset_id = dataset.id
        start_response = cls._start_usecase(project_id,
                                            name,
                                            dataset_id=dataset_id,
                                            data_type=data_type,
                                            training_type=training_type,
                                            metric=metric if isinstance(metric, str) else metric.value,
                                            **training_args)
        usecase = cls.from_id(start_response['_id'])
        events_url = '/{}/{}'.format(cls.resource, start_response['_id'])
        assert pio.client.event_manager is not None
        pio.client.event_manager.wait_for_event(usecase.resource_id,
                                                cls.resource,
                                                EventTuple('USECASE_VERSION_UPDATE', 'state', 'running',
                                                           [('state', 'failed')]),
                                                specific_url=events_url)

        return usecase

    def new_version(
        self,
        description: str = None,
        dataset: Union[Dataset, Tuple[Dataset, DatasetImages]] = None,
        column_config: ColumnConfig = None,
        metric: metrics.Enum = None,
        holdout_dataset: Dataset = None,
        training_config: TrainingConfig = None,
        **fit_params
    ) -> 'Supervised':
        """ Start a supervised usecase training to create a new version of the usecase (on the
        platform): the training configs are copied from the current version and then overridden
        for the given parameters.

        Args:
            description (str, optional): additional description of the version
            dataset (:class:`.Dataset`, :class:`.DatasetImages`, optional): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`, optional): Column configuration for the usecase
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (metrics.Enum, optional): Specific metric to use for the usecase (default: ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)
        Returns:
            :class:`.Supervised`: Newly created supervised usecase object (new version)
        """

        if column_config is None:
            column_config = self.column_config

        if training_config is None:
            training_config = self.training_config

        metric_str: str = self.metric
        if metric is not None:
            metric_str = metric.value

        if dataset is None:
            dataset_ids = self.dataset_id
        else:
            if isinstance(dataset, Dataset):
                dataset_ids = dataset.id
            elif isinstance(dataset, tuple):
                dataset_ids = [d.id for d in dataset]

        if holdout_dataset is None:
            holdout_dataset_id = self.holdout_dataset_id
        else:
            holdout_dataset_id = holdout_dataset.id

        params = {
            'dataset_id': dataset_ids,
            'metric': metric_str,
            'holdout_dataset': holdout_dataset_id,
            'training_type': self.training_type.value,
            'usecase_id': self._id,
            'parent_version': self.version,
            # 'nextVersion': max([v['version'] for v in self.versions]) + 1  FA: wait what ?
        }

        if description:
            params["description"] = description

        params.update(to_json(column_config))
        params.update(to_json(training_config))

        params.update(fit_params)
        endpoint = "/usecases/{}/versions".format(self.usecase_id)
        resp = client.request(endpoint=endpoint,
                              data=params,
                              method=requests.post,
                              content_type='application/json',
                              message_prefix='Usecase creation')
        json = parse_json(resp)

        usecase = self.from_id(json["_id"])

        events_url = '/{}/{}'.format(self.resource, json['_id'])
        assert client.event_manager is not None
        client.event_manager.wait_for_event(usecase.resource_id,
                                            self.resource,
                                            EventTuple('USECASE_VERSION_UPDATE', 'state', 'running',
                                                       [('state', 'failed')]),
                                            specific_url=events_url)

        return usecase

    def _save_json(self):
        json_dict = {
            '_id': self.id,
            'usecase_version_params': self._status.get('usecase_version_params', {}),
            'dataset_id': self._status['dataset_id'],
            'training_type': self.training_type.value,
            'data_type': self.data_type.value,
        }
        if self.holdout_dataset_id:
            json_dict['holdout_id'] = self.holdout_dataset_id
        return json_dict



# -*- coding: utf-8 -*-
from __future__ import print_function
import requests

from . import client
from .utils import parse_json, PrevisionException
from .api_resource import ApiResource, UniqueResourceMixin


class DataSource(ApiResource, UniqueResourceMixin):

    """ A datasource to access a distant data pool and create or fetch data easily. This
    resource is linked to a :class:`.Connector` resource that represents the connection to
    the distant data source.

    Args:
        _id (str): Unique id of the datasource
        connector (:class:`.Connector`): Reference to the associated connector (the resource
            to go through to get a data snapshot)
        name (str): Name of the datasource
        path (str, optional): Path to the file to fetch via the connector
        database (str, optional): Name of the database to fetch data from via the
            connector
        table (str, optional): Name of the table  to fetch data from via the connector
        request (str, optional): Direct SQL request to use with the connector to fetch data
    """

    resource = 'data-sources'

    def __init__(self, _id, connector_id: str, name: str, path: str = None, database: str = None,
                 table: str = None, request: str = None, gCloud=None, **kwargs):
        """ Instantiate a new :class:`.DataSource` object to manipulate a datasource resource
        on the platform. """
        super().__init__(_id=_id,
                         connector=connector_id,
                         name=name,
                         path=path,
                         database=database,
                         table=table,
                         request=request,
                         gCloud=gCloud)

        self._id = _id
        self.connector = connector_id

        self.name = name
        self.path = path
        self.database = database
        self.table = table
        self.request = request
        self.gCloud = gCloud

        self.other_params = kwargs

    @classmethod
    def list(cls, project_id: str, all: bool = False):
        """ List all the available datasources in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.DataSource` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.DataSource`): Fetched datasource objects
        """
        # FIXME : get /resource return type should be consistent
        resources = super()._list(all=all, project_id=project_id)
        return [cls(**source_data) for source_data in resources]

    @classmethod
    def from_id(cls, _id: str):
        """Get a datasource from the instance by its unique id.

        Args:
            _id (str): Unique id of the resource to retrieve

        Returns:
            :class:`.DataSource`: The fetched datasource

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing the result
        """
        # FIXME GET datasource should not return a dict with a "data" key
        url = '/{}/{}'.format(cls.resource, _id)
        resp = client.request(url, method=requests.get, message_prefix='From id data source')
        resp_json = parse_json(resp)

        return cls(**resp_json)

    @classmethod
    def _new(cls, project_id: str, connector, name: str, path: str = None, database: str = None, table: str = None,
             bucket=None, request=None, gCloud=None):
        """ Create a new datasource object on the platform.

        Args:
            connector (:class:`.Connector`): Reference to the associated connector (the resource
                to go through to get a data snapshot)
            name (str): Name of the datasource
            path (str, optional): Path to the file to fetch via the connector
            database (str, optional): Name of the database to fetch data from via the
                connector
            table (str, optional): Name of the table  to fetch data from via the connector
            request (str, optional): Direct SQL request to use with the connector to fetch data

        Returns:
            :class:`.DataSource`: The registered datasource object in the current workspace

        Raises:
            PrevisionException: Any error while uploading data to the platform
                or parsing the result
            Exception: For any other unknown error
        """

        data = {
            'connector_id': connector._id,
            'name': name,
            'path': path,
            'database': database,
            'bucket': bucket,
            'table': table,
            'request': request
        }
        if gCloud:
            data['g_cloud'] = gCloud

        url = '/projects/{}/{}'.format(project_id, cls.resource)
        resp = client.request(url,
                              data=data,
                              method=requests.post,
                              message_prefix='Datasource creation')
        json = parse_json(resp)

        if '_id' not in json:
            if 'message' in json:
                raise PrevisionException(json['message'])
            else:
                raise Exception('unknown error: {}'.format(json))
        return cls(json['_id'], connector, name, path, database, table, request)



from __future__ import print_function

import os
import copy
from typing import Dict, Union
import requests
import time
import json
import threading

from requests.models import Response

import previsionio.utils
from .logger import logger, event_logger
from . import config
from .utils import handle_error_response, parse_json, PrevisionException


class EventManager:
    def __init__(self, event_endpoint: str, auth_headers: Dict[str, str], client: 'Client'):
        self.event_endpoint = event_endpoint
        auth_headers = copy.deepcopy(auth_headers)
        self.headers = auth_headers
        self.client = client
        self.t = threading.Thread(target=self.update_events, daemon=True)
        self.t.start()

        self._events = (threading.Semaphore(1), {})

    @property
    def events(self):
        events_dict_copy = {}
        semd, event_dict = self._events

        with semd:
            for key, (semi, event_list) in event_dict.items():
                with semi:
                    events_dict_copy[key] = copy.deepcopy(event_list)

        return events_dict_copy

    def wait_for_event(self,
                       resource_id: str,
                       resource_type: str,
                       event_tuple: previsionio.utils.EventTuple,
                       specific_url: str):
        # ignore invalid resource ids
        if not isinstance(resource_id, str):
            return
        self.register_resource(resource_id)
        t0 = time.time()
        while time.time() < t0 + config.default_timeout:
            reconnect_start = time.time()
            semd, event_dict = self._events
            semd.acquire()
            if self.check_resource_event(resource_id, specific_url, event_tuple, semd):
                return
            semd.release()
            time.sleep(0.1)
            while time.time() < reconnect_start + 300:
                semd, event_dict = self._events
                semd.acquire()
                semi, event_list = event_dict[resource_id]
                remaining_events = []
                with semi:
                    for event in event_list:
                        if event.get('event') == event_tuple.name:
                            if self.check_resource_event(resource_id, specific_url, event_tuple, semd):
                                return
                        else:
                            remaining_events.append(event)

                event_dict[resource_id] = semi, remaining_events
                semd.release()
                time.sleep(0.1)
        else:
            raise TimeoutError('Failed to get status {} on {} {}'.format(event_tuple,
                                                                         resource_type,
                                                                         resource_id))

    def check_resource_event(self, resource_id: str, endpoint: str, event_tuple: previsionio.utils.EventTuple,
                             semd: threading.Semaphore):
        resp = self.client.request(endpoint=endpoint, method=requests.get, check_response=False)
        json_response = parse_json(resp)
        for k, v in event_tuple.fail_checks:
            if json_response.get(k) == v:
                semd.release()
                msg = 'Error on resource {}: {}\n{}'.format(resource_id,
                                                            json_response.get('errorMessage', ''),
                                                            json_response)
                raise PrevisionException(msg)
        if json_response.get(event_tuple.key) == event_tuple.value:
            semd.release()
            return True
        return False

    def register_resource(self, resource_id: str):
        event_logger.debug('Registering resource with id {}'.format(resource_id))
        payload = {'event': 'REGISTER', '_id': resource_id}
        self.add_event(resource_id, payload)

    def update_events(self):
        sse_timeout = 300
        event_data = None
        event_name = None

        while True:
            sse = requests.get(self.event_endpoint,
                               stream=True,
                               headers=self.headers,
                               timeout=sse_timeout)

            try:
                for line in sse.iter_lines(chunk_size=None):
                    if not line:
                        # filter out keep-alive new lines
                        continue

                    event_logger.debug('url: {} -- data: {}'.format(self.event_endpoint, line))
                    line = line.decode()
                    if line.startswith(':'):
                        # SSE comments can start with ":" character
                        event_logger.debug('sse comment{}'.format(line))
                        continue
                    elif line.startswith('id '):
                        # skip id message information
                        continue
                    elif line.startswith('event: '):
                        # get event name
                        event_name = line[len('event: '):]
                    elif line.startswith('data: '):
                        if event_name is None:
                            # ignore event with unknown name
                            continue

                        # get event data
                        try:
                            event_data = json.loads(line[len('data: '):])
                        except Exception as e:
                            # skip malformed event data
                            event_logger.warning('failed to parse json: "{}" -- error: {}'.format(line, e.__repr__()))
                            continue

                        resource_id = event_data.get('_id', None)

                        if not isinstance(resource_id, str):
                            # ignore event with invalid resource id
                            continue

                        # add event only if monitored resource
                        semd, event_dict = self._events
                        if resource_id in event_dict:
                            payload = {'event': event_name, 'id': resource_id}
                            event_logger.debug('url: {} -- event: {} payload: {}'.format(self.event_endpoint,
                                                                                         event_name,
                                                                                         payload))
                            self.add_event(resource_id, payload)
            except requests.exceptions.ChunkedEncodingError:
                event_logger.warning('closing connection to endpoint: "{}"'.format(self.event_endpoint))
            except requests.exceptions.ConnectionError:
                logger.warning('{}: no messages in {} seconds. reconnecting'.format(self.event_endpoint, sse_timeout))
            except Exception as e:
                logger.error(e)
            finally:
                sse.close()

    def add_event(self, resource_id: str, payload: Union[None, Dict]):
        event_logger.debug('adding event for {}:\npayload = {}'.format(resource_id, payload))
        semd, event_dict = self._events

        if payload and isinstance(payload, dict):
            with semd:
                if resource_id in event_dict:
                    semi, event_list = event_dict[resource_id]
                else:
                    semi, event_list = (threading.Semaphore(1), [])

                with semi:
                    event_list.append(payload)

                event_dict[resource_id] = semi, event_list

                self._events = semd, event_dict


class Client(object):

    """Client class to interact with the Prevision.io platform and manage authentication."""

    def __init__(self):
        self.token = None
        self.prevision_url = None
        self.user_info = None
        self.headers = {
            'accept-charset': 'UTF-8',
            'cache-control': 'no-cache',
            'accept': 'application/json',
        }

        self.api_version = '/ext/v1'

        self.url = None

        self.event_manager = None

    def _check_token_url(self):

        if not self.token:
            raise PrevisionException('No token configured. Call client.init_client() to initialize')

        if not self.prevision_url:
            raise PrevisionException('No url configured. Call client.init_client() to initialize')

    def request(self, endpoint: str, method, files: Dict = None, data: Dict = None,
                format: Dict = None, allow_redirects: bool = True,
                content_type: str = None, check_response: bool = True,
                message_prefix: str = None, **requests_kwargs) -> Response:
        """
        Make a request on the desired endpoint with the specified method & data.

        Requires initialization.

        Args:
            endpoint: (str): api endpoint (e.g. /usecases, /prediction/file)
            method (requests.{get,post,delete}): requests method
            files (dict): files dict
            data (dict): for single predict
            content_type (str): force request content-type
            allow_redirects (bool): passed to requests method
            check_response (bool): wether to handle error or not
            message_prefix (str): prefix message in error logs

        Returns:
            request response

        Raises:
            Exception: Error if url/token not configured
        """
        self._check_token_url()

        headers = copy.deepcopy(self.headers)
        if content_type:
            headers['content-type'] = content_type

        if self.url is None:
            raise RuntimeError("client.url not properly initialized")

        if format and len(format):
            def format_bool(val):
                if isinstance(val, bool):
                    if val:
                        return "true"
                    return "false"
                return val
            endpoint += '?' + '&'.join(["{}={}".format(key, format_bool(val)) for key, val in format.items()])

        url = self.url + endpoint

        status_code = 502
        retries = config.request_retries
        n_tries = 0
        resp = None
        while (n_tries < retries) and (status_code in config.retry_codes):
            n_tries += 1

            try:
                resp = method(url,
                              headers=headers,
                              files=files,
                              allow_redirects=allow_redirects,
                              json=data,
                              **requests_kwargs)
                status_code = resp.status_code

            except Exception as e:
                logger.warning(f'Failed to request {url} retrying {retries - n_tries} times: {e.__repr__()}')
                if n_tries == retries:
                    raise PrevisionException(f'Error requesting: {url} after {n_tries} retries')
                continue

            if status_code in config.retry_codes:
                time.sleep(config.request_retry_time)

        assert isinstance(resp, Response)
        if check_response:
            handle_error_response(resp=resp,
                                  url=url,
                                  data=data,
                                  files=files,
                                  message_prefix=message_prefix,
                                  n_tries=n_tries)

        return resp

    def update_user_info(self):
        assert self.url is not None
        user_info_response = requests.get(self.url + '/profile',
                                          headers=self.headers)
        result = parse_json(user_info_response)

        if 'err_code' in result and result['err_code'] == 'E_UNK':
            raise ValueError('Wrong token ' + str(result))
        self.user_info = result

    def init_client(self, prevision_url: str, token: str):
        """
        Init the client (and check that the connection is valid).

        Args:
            prevision_url (str): URL of the Prevision.io platform. Should be of the form
                https://<instance_name>.prevision.io, or a custom IP address if working on-premise.

            token (str): Your Prevision.io master token. Can be retrieved on
                /dashboard/infos on the web interface or obtained programmatically through:

                .. code-block:: python

                    client.init_client(prevision_url, token)
        """
        self.prevision_url = prevision_url
        self.url = self.prevision_url + self.api_version
        self.token = token
        self.headers['Authorization'] = self.token

        # check for correct connection
        resp = self.request('/version', requests.get)
        handle_error_response(resp, '/version')

        logger.debug('subscribing to events manager')
        self.event_manager = EventManager(self.url + '/events', auth_headers=self.headers, client=self)


client = Client()


if os.getenv('PREVISION_URL') and os.getenv('PREVISION_MASTER_TOKEN'):
    logger.info('Initializing Prevision.io client using environment variables')
    logger.debug('PREVISION_URL:' + os.getenv('PREVISION_URL', ""))
    logger.debug('PREVISION_MASTER_TOKEN:' + os.getenv('PREVISION_MASTER_TOKEN', ""))
    client.init_client(os.getenv('PREVISION_URL', ""), os.getenv('PREVISION_MASTER_TOKEN', ""))



from __future__ import print_function
import logging
from previsionio.logger import logger, event_logger


__version__ = '11.1.0'


def verbose(v, debug: bool = False, event_log: bool = False):
    """ Set the SDK level of verbosity.

    Args:
        v (bool): Whether to activate info logging
        debug (bool, optional): Whether to activate detailed
            debug logging (default: ``False``)
        event_log (bool, optional): Whether to activate detailed
            event managers debug logging (default: ``False``)
    """
    if event_log:
        event_logger.setLevel(logging.DEBUG)

    if v:
        if debug:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)


verbose(False)


class Config:
    def __init__(self):
        self.auto_update: bool = True
        self.zip_files: bool = True
        self.success_codes: list = list(range(200, 211)) + [226]
        self.retry_codes: list = [500, 502, 503, 504]
        self.request_retries: int = 6
        self.request_retry_time: int = 10
        self.scheduler_refresh_rate: int = 10
        self.default_timeout: float = 3600.


config = Config()

from previsionio.prevision_client import client
from previsionio.usecase_config import (
    AdvancedModel,
    NormalModel,
    SimpleModel,
    TypeProblem,
    Feature,
    TrainingConfig,
    Profile,
    base_config,
    quick_config,
    ultra_config,
    ColumnConfig,
    YesOrNo,
    YesOrNoOrAuto,
)

import previsionio.metrics as metrics

from previsionio.connector import Connector
from previsionio.datasource import DataSource
from previsionio.project import Project
from previsionio.usecase import Usecase
from previsionio.supervised import Supervised
from previsionio.timeseries import (
    TimeSeries,
    TimeWindow,
    TimeWindowException,
)
from previsionio.model import (
    Model,
    ClassificationModel,
    RegressionModel,
    MultiClassificationModel,
    TextSimilarityModel,
)
from previsionio.text_similarity import (
    TextSimilarity,
    DescriptionsColumnConfig,
    QueriesColumnConfig,
    ListModelsParameters,
    ModelsParameters,
    TextSimilarityModels,
    ModelEmbedding,
    Preprocessing,
)
from previsionio.dataset import Dataset, DatasetImages
from previsionio.analyzer import cv_classif_analysis
from previsionio.deployed_model import DeployedModel
from previsionio.usecase_deployment import UsecaseDeployment
from previsionio.prediction import ValidationPrediction, DeploymentPrediction

__all__ = [
    'client',
    'AdvancedModel',
    'NormalModel',
    'SimpleModel',
    'TypeProblem',
    'Feature',
    'TrainingConfig',
    'YesOrNo',
    'YesOrNoOrAuto',
    'Profile',
    'base_config',
    'quick_config',
    'ultra_config',
    'ColumnConfig',
    'metrics',
    'Connector',
    'DataSource',
    'Supervised',
    'TimeSeries',
    'TimeWindow',
    'TimeWindowException',
    'Dataset',
    'DatasetImages',
    'Model',
    'RegressionModel',
    'ClassificationModel',
    'MultiClassificationModel',
    'TextSimilarityModel',
    'cv_classif_analysis',
    'DeployedModel',
    'TextSimilarity',
    'DescriptionsColumnConfig',
    'QueriesColumnConfig',
    'ListModelsParameters',
    'ModelsParameters',
    'TextSimilarityModels',
    'ModelEmbedding',
    'Preprocessing',
    'Project',
    'Usecase',
    'UsecaseDeployment',
    'ValidationPrediction',
    'DeploymentPrediction'
]



# -*- coding: utf-8 -*-
from __future__ import print_function
from enum import Enum
from typing import Dict, Tuple

from requests.models import Response
from previsionio import metrics
from previsionio.usecase_config import ColumnConfig, DataType, TrainingConfig, TypeProblem
import requests

from . import client
from .utils import parse_json, PrevisionException

from .api_resource import ApiResource, UniqueResourceMixin
from .datasource import DataSource
from .dataset import Dataset, DatasetImages
from .connector import Connector, SQLConnector, FTPConnector, \
    SFTPConnector, S3Connector, HiveConnector, GCPConnector
from .supervised import Supervised
from .timeseries import TimeSeries, TimeWindow
from .text_similarity import (DescriptionsColumnConfig, ListModelsParameters, QueriesColumnConfig,
                              TextSimilarity, TextSimilarityLang)
from .usecase import Usecase
from .usecase_deployment import UsecaseDeployment
from pandas import DataFrame


class ProjectColor(Enum):
    greenLighter = '#16D92C'
    green = '#29CC3C'
    greenDarker = '#47B354'
    greenDarkest = '#5C9963'
    purpleLighter = '#8E00E6'
    purple = '#8D29CC'
    purpleDarker = '#8947B3'
    purpleDarkest = '#825C99'
    blueLighter = '#0099E6'
    blue = '#2996CC'
    blueDarker = '#478FB3'
    blueDarkest = '#5C8599'
    yellowLighter = '#E6E600'
    yellow = '#CCCC29'
    yellowDarker = '#B3B347'
    yellowDarkest = '#99995C'


class Project(ApiResource, UniqueResourceMixin):

    """ A Project

    Args:
        _id (str): Unique id of the project
        name (str): Name of the project
        description(str, optional): Description of the project
        color (ProjectColor, optional): Color of the project

    """

    resource = 'projects'

    def __init__(self, _id: str, name: str, description: str = None, color: ProjectColor = None, created_by: str = None,
                 admins=[], contributors=[], viewers=[], pipelines_count: int = 0, usecases_count: int = 0,
                 dataset_count: int = 0, **kwargs):
        """ Instantiate a new :class:`.Project` object to manipulate a project resource
        on the platform. """
        super().__init__(_id=_id,
                         name=name,
                         description=description,
                         color=color)

        self._id = _id
        self.name = name
        self.description = description
        self.color = color
        self.created_by = created_by
        self.admins = admins
        self.contributors = contributors
        self.viewers = viewers
        self.pipelines_count = pipelines_count
        self.usecases_count = usecases_count
        self.dataset_count = dataset_count

    @classmethod
    def list(cls, all: bool = False):
        """ List all the available project in the current active [client] workspace.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.
        Returns:
            list(:class:`.Project`): Fetched project objects
        """
        # FIXME : get /resource return type should be consistent
        resources = super()._list(all=all)
        return [cls(**source_data) for source_data in resources]

    @classmethod
    def from_id(cls, _id: str):
        """Get a project from the instance by its unique id.

        Args:
            _id (str): Unique id of the resource to retrieve

        Returns:
            :class:`.Project`: The fetched datasource

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing the result
        """
        # FIXME GET datasource should not return a dict with a "data" key
        url = '/{}/{}'.format(cls.resource, _id)
        resp = client.request(url,
                              method=requests.get,
                              message_prefix='Projects list')
        resp_json = parse_json(resp)

        return cls(**resp_json)

    @property
    def id(self):
        return self._id

    def users(self):
        """Get a project from the instance by its unique id.

        Args:
            _id (str): Unique id of the resource to retrieve

        Returns:
            :class:`.Project`: The fetched datasource

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing the result
        """

        end_point = '/{}/{}/users'.format(self.resource, self._id)
        response = client.request(endpoint=end_point,
                                  method=requests.get,
                                  message_prefix='Fetching user for project id {}'.format(self._id))

        res = parse_json(response)
        return res

    def info(self) -> Dict:
        """Get a datasource from the instance by its unique id.

        Args:
            _id (str): Unique id of the resource to retrieve

        Returns:
            Dict: Information about the Project with these entries:
                "_id"
                "name"
                "description"
                "color"
                "created_by"
                "admins"
                "contributors"
                "viewers"
                "pipelines_count"
                "usecases_count"
                "dataset_count"
                "users"

        Raises:
            PrevisionException: Any error while fetching data from the platform
                or parsing the result
        """
        project_info = {"_id": self._id,
                        "name": self.name,
                        "description": self.description,
                        "color": self.color,
                        "created_by": self.created_by,
                        "admins": self.admins,
                        "contributors": self.contributors,
                        "viewers": self.viewers,
                        "pipelines_count": self.pipelines_count,
                        "usecases_count": self.usecases_count,
                        "dataset_count": self.dataset_count,
                        "users": self.users}
        return project_info

    @classmethod
    def new(cls, name: str, description: str = None, color: ProjectColor = None) -> 'Project':
        """ Create a new project on the platform.

        Args:
            name (str): Name of the project
            description(str, optional): Description of the project
            color (str, optional): Color of the project

        Returns:
            :class:`.Project`: The registered project object in the current workspace

        Raises:
            PrevisionException: Any error while uploading data to the platform
                or parsing the result
            Exception: For any other unknown error
        """

        data = {
            'name': name,
            'description': description
        }
        if color:
            data['color'] = color.value

        url = '/{}'.format(cls.resource)
        resp = client.request(url,
                              data=data,
                              method=requests.post,
                              message_prefix='Project creation')

        json = parse_json(resp)

        if '_id' not in json:
            if 'message' in json:
                raise PrevisionException(json['message'])
            else:
                raise Exception('unknown error: {}'.format(json))

        return cls(json['_id'], name, description, color, json['created_by'],
                   json['admins'], json['contributors'], json['pipelines_count'])

    def delete(self) -> Response:
        """Delete a project from the actual [client] workspace.

        Raises:
            PrevisionException: If the dataset does not exist
            requests.exceptions.ConnectionError: Error processing the request
        """
        resp = client.request(endpoint='/{}/{}'.format(self.resource, self.id),
                              method=requests.delete,
                              message_prefix='Project delete')
        return resp

    def create_dataset(self, name: str, datasource: DataSource = None, file_name: str = None,
                       dataframe: DataFrame = None, **kwargs):
        """ Register a new dataset in the workspace for further processing.
        You need to provide either a datasource, a file name or a dataframe
        (only one can be specified).

        .. note::

            To start a new use case on a dataset, it has to be already
            registred in your workspace.

        Args:
            name (str): Registration name for the dataset
            datasource (:class:`.DataSource`, optional): A DataSource object used
                to import a remote dataset (if you want to import a specific dataset
                from an existent database, you need a datasource connector
                (:class:`.Connector` object) designed to point to the related data source)
            file_name (str, optional): Path to a file to upload as dataset
            dataframe (pd.DataFrame, optional): A ``pandas`` dataframe containing the
                data to upload

        Raises:
            Exception: If more than one of the keyword arguments ``datasource``, ``file_name``,
                ``dataframe`` was specified
            PrevisionException: Error while creating the dataset on the platform

        Returns:
            :class:`.Dataset`: The registered dataset object in the current workspace.
        """
        return Dataset._new(self._id, name, datasource=datasource, file_name=file_name, dataframe=dataframe, **kwargs)

    def list_datasets(self, all: bool = True):
        """ List all the available datasets in the current active project.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.Dataset` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.Dataset`): Fetched dataset objects
        """
        return Dataset.list(self._id, all=all)

    def create_image_folder(self, name: str, file_name: str):
        """ Register a new image dataset in the workspace for further processing
        (in the image folders group).

        .. note::

            To start a new use case on a dataset image, it has to be already
            registred in your workspace.

        Args:
            name (str): Registration name for the dataset
            file_name (str): Path to the zip file to upload as image dataset

        Raises:
            PrevisionException: Error while creating the dataset on the platform

        Returns:
            :class:`.DatasetImages`: The registered dataset object in the current workspace.
        """
        return DatasetImages._new(self._id, name, file_name)

    def list_image_folders(self, all: bool = True):
        """ List all the available dataset image in the current active [client] workspace.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.DatasetImages` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.DatasetImages`): Fetched dataset objects
        """
        return DatasetImages.list(self._id, all=all)

    def create_sql_connector(self, name: str, host: str, port: int = 3306, username: str = '', password: str = ''):
        """ A connector to interact with a distant source of data (and
        easily get data snapshots using an associated :class:`.DataSource`
        resource).

        Args:
            name (str): Name of the connector
            host (str): Url of the connector
            port (int): Port of the connector
            username (str, optional): Username to use connect to the remote data source
            password (str, optional): Password to use connect to the remote data source
        Returns:
            :class:`.SQLConnector`: The registered connector object in the current project.
        """
        return SQLConnector._new(self._id, name, host, port, 'SQL', username=username, password=password)

    def create_ftp_connector(self, name: str, host: str, port: int = 21, username: str = '', password: str = ''):
        """ A connector to interact with a distant source of data (and
        easily get data snapshots using an associated :class:`.DataSource`
        resource).

        Args:
            name (str): Name of the connector
            host (str): Url of the connector
            port (int): Port of the connector
            username (str, optional): Username to use connect to the remote data source
            password (str, optional): Password to use connect to the remote data source
        Returns:
            :class:`.FTPConnector`: The registered connector object in the current project.
        """
        return FTPConnector._new(self._id, name, host, port, 'FTP', username=username, password=password)

    def create_sftp_connector(self, name: str, host: str, port: int = 23, username: str = '', password: str = ''):
        """ A connector to interact with a distant source of data (and
        easily get data snapshots using an associated :class:`.DataSource`
        resource).

        Args:
            name (str): Name of the connector
            host (str): Url of the connector
            port (int): Port of the connector
            username (str, optional): Username to use connect to the remote data source
            password (str, optional): Password to use connect to the remote data source
        Returns:
            :class:`.SFTPConnector`: The registered connector object in the current project.
        """
        return SFTPConnector._new(self._id, name, host, port, 'SFTP', username=username, password=password)

    def create_s3_connector(self, name: str, host: str = '', port: int = None, username: str = '', password: str = ''):
        """ A connector to interact with a distant source of data (and
        easily get data snapshots using an associated :class:`.DataSource`
        resource).

        Args:
            name (str): Name of the connector
            host (str): Url of the connector
            port (int): Port of the connector
            username (str, optional): Username to use connect to the remote data source
            password (str, optional): Password to use connect to the remote data source
        Returns:
            :class:`.S3Connector`: The registered connector object in the current project.
        """
        return S3Connector._new(self._id, name, host, port, 'S3', username=username, password=password)

    def create_hive_connector(self, name: str, host: str, port: int = 10000, username: str = '', password: str = ''):
        """ A connector to interact with a distant source of data (and
        easily get data snapshots using an associated :class:`.DataSource`
        resource).

        Args:
            name (str): Name of the connector
            host (str): Url of the connector
            port (int): Port of the connector
            username (str, optional): Username to use connect to the remote data source
            password (str, optional): Password to use connect to the remote data source
        Returns:
            :class:`.HiveConnector`: The registered connector object in the current project.
        """
        return HiveConnector._new(self._id, name, host, port, 'HIVE', username=username, password=password)

    def create_gcp_connector(self, name: str = '', host: str = '', port=None, username: str = '', password: str = '',
                             googleCredentials: str = ''):
        """ A connector to interact with a distant source of data (and
        easily get data snapshots using an associated :class:`.DataSource`
        resource).

        Args:
            name (str): Name of the connector
            googleCredentials(str): google credentials
        Returns:
            :class:`.GCPConnector`: The registered connector object in the current project.
        """
        return GCPConnector._new(self._id, name, host, port, 'GCP', username=username, password=password,
                                 googleCredentials=googleCredentials)

    def list_connectors(self, all: bool = True):
        """ List all the available connectors in the current active project.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.Connector` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.Connector`): Fetched dataset objects
        """
        return Connector.list(self._id, all=all)

    def create_datasource(self, connector: Connector, name: str, path: str = None, database: str = None,
                          table: str = None, bucket: str = None, request: str = None, gCloud: str = None):
        """ Create a new datasource object on the platform.

        Args:
            connector (:class:`.Connector`): Reference to the associated connector (the resource
                to go through to get a data snapshot)
            name (str): Name of the datasource
            path (str, optional): Path to the file to fetch via the connector
            database (str, optional): Name of the database to fetch data from via the
                connector
            table (str, optional): Name of the table to fetch data from via the connector
            bucket (str, optional): Name of the bucket to fetch data from via the connector
            gCloud (str, optional): gCloud
            request (str, optional): Direct SQL request to use with the connector to fetch data
        Returns:
            :class:`.DataSource`: The registered datasource object in the current project

        Raises:
            PrevisionException: Any error while uploading data to the platform
                or parsing the result
            Exception: For any other unknown error
        """
        return DataSource._new(self._id, connector, name, path=path, database=database,
                               table=table, bucket=bucket, request=request, gCloud=gCloud)

    def list_datasource(self, all: bool = False):
        """ List all the available datasources in the current active project.

        .. warning::

            Contrary to the parent ``list()`` function, this method
            returns actual :class:`.DataSource` objects rather than
            plain dictionaries with the corresponding data.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.

        Returns:
            list(:class:`.DataSource`): Fetched dataset objects
        """
        return DataSource.list(self._id, all=all)

    def fit_regression(self, name: str, dataset: Dataset, column_config: ColumnConfig,
                       metric: metrics.Regression = metrics.Regression.RMSE, holdout_dataset=None,
                       training_config=TrainingConfig(), **kwargs):
        """ Start a tabular regression usecase version training

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`): Column configuration for the usecase
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (str, optional): Specific metric to use for the usecase (default: ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)

        Returns:
            :class:`.supervised.Regression`: Newly created Regression usecase version object
        """
        return Supervised._fit(
            self._id, name,
            data_type=DataType.Tabular,
            training_type=TypeProblem.Regression,
            dataset=dataset,
            column_config=column_config,
            metric=metric,
            holdout_dataset=holdout_dataset,
            training_config=training_config,
            **kwargs
        )

    def fit_classification(self, name: str, dataset: Dataset, column_config: ColumnConfig,
                           metric: metrics.Classification = metrics.Classification.AUC, holdout_dataset=None,
                           training_config=TrainingConfig(), **kwargs):
        """ Start a tabular classification usecase version training

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`): Column configuration for the usecase
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (str, optional): Specific metric to use for the usecase (default: ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)

        Returns:
            :class:`.supervised.Classification`: Newly created Classification usecase version object
        """
        return Supervised._fit(
            self._id, name,
            data_type=DataType.Tabular,
            training_type=TypeProblem.Classification,
            dataset=dataset,
            column_config=column_config,
            metric=metric,
            holdout_dataset=holdout_dataset,
            training_config=training_config,
            **kwargs
        )

    def fit_multiclassification(self, name: str, dataset: Dataset, column_config: ColumnConfig,
                                metric: metrics.MultiClassification = metrics.MultiClassification.log_loss,
                                holdout_dataset: Dataset = None, training_config=TrainingConfig(), **kwargs):
        """ Start a tabular multiclassification usecase version training

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`): Column configuration for the usecase
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (str, optional): Specific metric to use for the usecase (default: ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)

        Returns:
            :class:`.supervised.MultiClassification`: Newly created MultiClassification usecase version object
        """
        return Supervised._fit(
            self._id, name,
            data_type=DataType.Tabular,
            training_type=TypeProblem.MultiClassification,
            dataset=dataset,
            column_config=column_config,
            metric=metric,
            holdout_dataset=holdout_dataset,
            training_config=training_config,
            **kwargs
        )

    def fit_image_regression(self, name: str, dataset: Tuple[Dataset, DatasetImages], column_config: ColumnConfig,
                             metric: metrics.Regression = metrics.Regression.RMSE, holdout_dataset: Dataset = None,
                             training_config=TrainingConfig(), **kwargs):
        """ Start an image regression usecase version training

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`, :class:`.DatasetImages`): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`): Column configuration for the usecase
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (str, optional): Specific metric to use for the usecase (default: ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)

        Returns:
            :class:`.supervised.RegressionImages`: Newly created RegressionImages usecase version object
        """
        return Supervised._fit(
            self._id, name,
            data_type=DataType.Images,
            training_type=TypeProblem.Regression,
            dataset=dataset,
            column_config=column_config,
            metric=metric,
            holdout_dataset=holdout_dataset,
            training_config=training_config,
            **kwargs
        )

    def fit_image_classification(self, name: str, dataset: Tuple[Dataset, DatasetImages], column_config: ColumnConfig,
                                 metric: metrics.Classification = metrics.Classification.AUC,
                                 holdout_dataset: Dataset = None, training_config=TrainingConfig(), **kwargs):
        """ Start an image classification usecase version training

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`, :class:`.DatasetImages`): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`): Column configuration for the usecase
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (str, optional): Specific metric to use for the usecase (default: ``None``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)

        Returns:
            :class:`.supervised.ClassificationImages`: Newly created ClassificationImages usecase version object
        """
        return Supervised._fit(
            self._id, name,
            data_type=DataType.Images,
            training_type=TypeProblem.Classification,
            dataset=dataset,
            column_config=column_config,
            metric=metric,
            holdout_dataset=holdout_dataset,
            training_config=training_config,
            **kwargs
        )

    def fit_image_multiclassification(
        self, name: str,
        dataset: Tuple[Dataset, DatasetImages],
        column_config: ColumnConfig,
        metric: metrics.MultiClassification = metrics.MultiClassification.log_loss,
        holdout_dataset: Dataset = None,
        training_config=TrainingConfig(),
        **kwargs
    ) -> Supervised:
        """ Start an image multiclassification usecase version training

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`, :class:`.DatasetImages`): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`): Column configuration for the usecase
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            metric (:enum: `metrics.MultiClassification`, optional): Specific metric to use for the usecase (default:
                ``metrics.MultiClassification.log_loss``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)

        Returns:
            :class:`.supervised.MultiClassificationImages`: Newly created MultiClassificationImages usecase version
                object
        """
        return Supervised._fit(
            self._id, name,
            data_type=DataType.Images,
            training_type=TypeProblem.MultiClassification,
            dataset=dataset,
            column_config=column_config,
            metric=metric,
            holdout_dataset=holdout_dataset,
            training_config=training_config,
            **kwargs
        )

    def fit_timeseries_regression(
        self,
        name: str,
        dataset: Dataset,
        column_config: ColumnConfig,
        time_window: TimeWindow,
        metric: metrics.Regression = metrics.Regression.RMSE,
        holdout_dataset: Dataset = None,
        training_config=TrainingConfig()
    ) -> TimeSeries:
        """ Start a timeseries regression usecase version training

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`): Reference to the dataset
                object to use for as training dataset
            column_config (:class:`.ColumnConfig`): Column configuration for the usecase version
                (see the documentation of the :class:`.ColumnConfig` resource for more details
                on each possible column types)
            time_window (:class:`.TimeWindow`): Time configuration
                (see the documentation of the :class:`.TimeWindow` resource for more details)
            metric (:enum: `metrics.Regression`, optional): Specific metric to use for the usecase (default:
                ``metrics.Regression.RMSE``)
            holdout_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a holdout dataset (default: ``None``)
            training_config (:class:`.TrainingConfig`): Specific training configuration
                (see the documentation of the :class:`.TrainingConfig` resource for more details
                on all the parameters)

        Returns:
            :class:`.TimeSeries`: Newly created TimeSeries usecase version object
        """
        return TimeSeries._fit(self._id, name, dataset, column_config, time_window, metric=metric,
                               holdout_dataset=holdout_dataset, training_config=training_config)

    def fit_text_similarity(self, name: str, dataset: Dataset, description_column_config: DescriptionsColumnConfig,
                            metric: metrics.TextSimilarity = metrics.TextSimilarity.accuracy_at_k, top_k: int = 10,
                            lang: TextSimilarityLang = TextSimilarityLang.Auto, queries_dataset: Dataset = None,
                            queries_column_config: QueriesColumnConfig = None,
                            models_parameters: ListModelsParameters = ListModelsParameters()):
        """ Start a text similarity usecase training with a specific training configuration.

        Args:
            name (str): Name of the usecase to create
            dataset (:class:`.Dataset`): Reference to the dataset
                object to use for as training dataset
            description_column_config (:class:`.DescriptionsColumnConfig`): Description column configuration
                (see the documentation of the :class:`.DescriptionsColumnConfig` resource for more details
                on each possible column types)
            metric (str, optional): Specific metric to use for the usecase (default: ``accuracy_at_k``)
            top_k (int, optional): top_k (default: ``10``)
            queries_dataset (:class:`.Dataset`, optional): Reference to a dataset object to
                use as a queries dataset (default: ``None``)
            queries_column_config (:class:`.QueriesColumnConfig`): Queries column configuration
                (see the documentation of the :class:`.QueriesColumnConfig` resource for more details
                on each possible column types)
            models_parameters (:class:`.ListModelsParameters`): Specific training configuration
                (see the documentation of the :class:`.ListModelsParameters` resource for more details
                on all the parameters)

        Returns:
            :class:`.previsionio.text_similarity.TextSimilarity`: Newly created TextSimilarity usecase version object
        """
        return TextSimilarity._fit(
            self._id,
            name,
            dataset,
            description_column_config,
            metric=metric,
            top_k=top_k,
            lang=lang,
            queries_dataset=queries_dataset,
            queries_column_config=queries_column_config,
            models_parameters=models_parameters
        )

    def list_usecases(self, all: bool = True):
        """ List all the available usecase in the current project.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.
        Returns:
            list(:class:`.Usecase`): Fetched usecase objects
        """
        return Usecase.list(self._id, all=all)

    def create_usecase_deployment(self, name: str, main_model, challenger_model=None, access_type: str = 'public'):
        return UsecaseDeployment._new(
            self._id,
            name,
            main_model,
            challenger_model=challenger_model,
            access_type=access_type
        )

    def list_usecase_deployments(self, all: bool = True):
        """ List all the available usecase in the current project.

        Args:
            all (boolean, optional): Whether to force the SDK to load all items of
                the given type (by calling the paginated API several times). Else,
                the query will only return the first page of result.
        Returns:
            list(:class:`.UsecaseDeployment`): Fetched usecase deployment objects
        """
        return UsecaseDeployment.list(self._id, all=all)


connectors_names = {
    'SQL': "create_sql_connector",
    'FTP': "create_ftp_connector",
    'SFTP': "create_sftp_connector",
    'S3': "create_s3_connector",
    'HIVE': "create_hive_connector",
    'GCP': "create_gcp_connector"
}



import sys
from setuptools import setup, find_packages


python_major = sys.version_info[0]

install_requires = [
    "pandas>=0.25.3",
    "requests>=2.23.0",
    "scikit-learn>=0.22.2",
    "numpy>=1.18.4",
    "requests-oauthlib>=1.3.0",
    "oauthlib>=3.1.0",
    "matplotlib>=3.2.1",
]

if python_major == 2:
    install_requires.append('enum')


with open("README.md") as readme:
    setup(name='previsionio',
          version='11.1.4',
          description='python wrapper for prevision api',
          long_description=readme.read(),
          url='https://github.com/previsionio/prevision-python',
          author='prevision.io',
          author_email='team-core@prevision.io',
          packages=find_packages(exclude=['ci*', 'utest*']),
          license="MIT License",
          install_requires=install_requires,
          zip_safe=False,
          python_requires=">= 3.6",
          keywords='ml, ai, prevision, sdk',
          classifier=[
              'Intended Audience :: Developers',
              'Natural Language :: English',
              'License :: OSI Approved :: MIT License',
              'Programming Language :: Python :: 3',
              'Programming Language :: Python :: 3.6',
              'Programming Language :: Python :: 3.7',
              'Programming Language :: Python :: 3.8',
              'Programming Language :: Python :: 3.9',
          ]
          )



import itertools
from previsionio.usecase_version import ClassicUsecaseVersion
from previsionio.usecase_config import TypeProblem
import numpy as np
import os

try:
    # the package is running inside the prevision.io notebook
    local_url = os.environ['PREVISION_WIDGET_URL']
except KeyError:
    # the package is running inside the prevision.io notebook
    from matplotlib import pyplot as plt
else:
    import IPython

from sklearn.metrics import roc_curve, auc, confusion_matrix, recall_score, f1_score, precision_score


class Plotter:
    def __init__(self):
        pass


class PrevisionioPlotter(Plotter):
    prevision_widgets = ['featdetail', 'featimp', 'regscores', 'multiclassifRoc', 'multiclassifConfusion',
                         'multiclassifScore', 'decisionChart', 'classifMetrics', 'classifDensity', 'classifRoc',
                         'clusterStats', 'costMatrix', 'confusionMatrix', 'regvs', 'regdisp', 'repartition',
                         'classifLift', 'classifPerBinLift']

    def __init__(self, usecase):
        super().__init__()

        def add_widget_method(widget):
            def f(plotter):
                return plotter.__plot(widget, usecase)

            setattr(PrevisionioPlotter, widget, f)

        for method in self.prevision_widgets:
            add_widget_method(method)

    def __plot(self, widget, usecase, width=1000, height=600):
        iframe_url = '<iframe src={}/widgets/{}/{} width={} height={}></iframe>'.format(local_url,
                                                                                        widget,
                                                                                        usecase.usecase_name,
                                                                                        height,
                                                                                        width)
        return IPython.display.HTML(iframe_url)


class PlotlyPlotter(Plotter):
    def __init__(self, usecase: ClassicUsecaseVersion):
        self.usecase = usecase
        super().__init__()

    def plot_roc(self):
        """
        Plot a ROC curve. Needs a prediction made with the truth present in the dataset.

        Only available for binary classification

        """

        import plotly.graph_objs as go
        from plotly.offline import iplot, init_notebook_mode
        from plotly import tools

        init_notebook_mode()

        if self.usecase.training_type not in [TypeProblem.Classification, TypeProblem.MultiClassification]:
            raise Exception(
                'ROC curve only available for classification or multiclassif, '
                'not ' + self.usecase.training_type.value)

        preds = self.usecase.get_cv()

        target_col_name = self.usecase.column_config.target_column
        assert target_col_name

        if self.usecase.training_type == TypeProblem.Classification:
            fpr, tpr, _ = roc_curve(preds[target_col_name], preds['pred_' + target_col_name])
            roc_auc = auc(fpr, tpr)

            lw = 1

            trace1 = go.Scatter(x=fpr, y=tpr,
                                mode='lines',
                                line=dict(color='darkorange', width=lw),
                                name='ROC curve (area = %0.2f)' % roc_auc
                                )

            trace2 = go.Scatter(x=[0, 1], y=[0, 1],
                                mode='lines',
                                line=dict(color='navy', width=.2, dash='dash'),
                                showlegend=False)

            layout = go.Layout(title='Receiver operating characteristic example',
                               xaxis=dict(title='False Positive Rate'),
                               yaxis=dict(title='True Positive Rate'))

            fig = go.Figure(data=[trace1, trace2], layout=layout)
            iplot(fig)

        else:
            max_cols = 3
            max_rows = 3

            labels = sorted(preds[target_col_name].unique())

            if len(labels) > 12:
                raise Exception('cannot plot roc curves for more than 12 classes')

            n_cols = min(max_cols, len(labels))
            n_rows = (len(labels) // (max_rows - 1))

            fig = tools.make_subplots(rows=n_rows,
                                      cols=n_cols,
                                      subplot_titles=['Class: {}'.format(label) for label in labels])

            for i, label in enumerate(labels):
                truth = preds[target_col_name] == label

                fpr, tpr, _ = roc_curve(truth, preds['pred_{}_{}'.format(target_col_name, label)])
                roc_auc = auc(fpr, tpr)

                col = i % max_cols
                row = i // max_rows

                lw = 1

                trace1 = go.Scatter(x=fpr, y=tpr,
                                    mode='lines',
                                    line=dict(color='darkorange', width=lw),
                                    name='ROC curve (area = {:0.2f})'.format(roc_auc))

                trace2 = go.Scatter(x=[0, 1], y=[0, 1],
                                    mode='lines',
                                    line=dict(color='navy', width=.2, dash='dash'),
                                    showlegend=False)

                fig.add_trace(trace1, row + 1, col + 1)
                fig.add_trace(trace2, row + 1, col + 1)

            fig['layout'].update(height=500 * n_rows, width=500 * n_cols, title='Multiclass ROC')
            iplot(fig)


class MatplotlibPlotter(Plotter):
    def __init__(self, usecase):
        self.usecase = usecase
        super().__init__()

    def featdetail(self):
        raise NotImplementedError

    def featimp(self):
        raise NotImplementedError

    def regscores(self):
        raise NotImplementedError

    def multiclassifRoc(self):
        raise NotImplementedError

    def multiclassifConfusion(self):
        raise NotImplementedError

    def multiclassifScore(self):
        raise NotImplementedError

    def decisionChart(self):
        raise NotImplementedError

    def classifMetrics(self):
        raise NotImplementedError

    def classifDensity(self):
        raise NotImplementedError

    def classifRoc(self):
        raise NotImplementedError

    def clusterStats(self):
        raise NotImplementedError

    def costMatrix(self):
        raise NotImplementedError

    def confusionMatrix(self):
        raise NotImplementedError

    def regvs(self):
        raise NotImplementedError

    def regdisp(self):
        raise NotImplementedError

    def repartition(self):
        raise NotImplementedError

    def classifLift(self):
        raise NotImplementedError

    def classifPerBinLift(self):
        raise NotImplementedError

    def plot_roc(self, predict_id=None):
        """
        Plot a ROC curve. Needs a prediction made with the truth present in the dataset.

        Only available for binary classification

        Args:
            usecase:
            predict_id (str): ID of the prediction
        """
        if self.usecase.training_type not in [TypeProblem.Classification, TypeProblem.MultiClassification]:
            raise Exception(
                'ROC curve only available for classification or multiclassif, '
                'not ' + self.usecase.training_type)

        if predict_id:
            raise NotImplementedError
            # if predict_8id not in self.usecase.predictions:
            #     raise KeyError('No such prediction')
            #
            # prediction = self.usecase.predictions[predict_id]
            #
            # if prediction.truth is None:
            #     raise ValueError('Truth not set for this prediction')
            # else:
            #     truth, pred = prediction.truth, prediction.data[self.usecase.usecase_params['target_column']]
        else:
            preds = self.usecase.get_cv()

        target_col_name = self.usecase.column_config.target_column

        if not plt:
            raise Exception('matplotlib not installed. please install to plot curves')

        lw = 1

        if self.usecase.training_type == TypeProblem.Classification:
            fpr, tpr, _ = roc_curve(preds[target_col_name], preds['pred_' + target_col_name])
            roc_auc = auc(fpr, tpr)
            lw = 2
            plt.figure()
            plt.plot(fpr, tpr, color='darkorange', lw=lw, label='AUC = {})'.format(roc_auc))
            plt.plot([0, 1], [0, 1], color='navy', lw=lw)
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title('ROC curve')
            plt.legend(loc='lower right')

        else:
            max_cols = 3
            max_rows = 3

            labels = preds[target_col_name].unique()

            if len(labels) > 9:
                raise Exception('cannot plot roc curves for more than 9 classes')

            n_cols = min(max_cols, len(labels))
            n_rows = (len(labels) // (max_rows - 1))

            f, axarr = plt.subplots(n_rows, n_cols)

            for i, label in enumerate(labels):
                truth = preds[target_col_name] == label

                fpr, tpr, _ = roc_curve(truth, preds['pred_{}_{}'.format(target_col_name, label)])
                roc_auc = auc(fpr, tpr)

                col = i % max_cols
                row = i // max_rows

                ax = axarr[row, col] if n_rows > 1 else axarr[col]

                ax.plot(fpr, tpr, color='darkorange', lw=lw, label='AUC = {})'.format(roc_auc))
                ax.plot([0, 1], [0, 1], color='navy', lw=lw)
                ax.set_xlim([0.0, 1.0])
                ax.set_ylim([0.0, 1.05])
                ax.set_xlabel('False Positive Rate')
                ax.set_ylabel('True Positive Rate')
                ax.legend(loc='lower right')

            plt.title('ROC curve')

        plt.show()

    def plot_conf_mat(self, predict_id):
        """
        Plot a confusion matrix. Needs a prediction made with the truth present in the dataset.

        Only available for multi-classification.

        Args:
            predict_id (str): ID of the prediction

        """
        if self.usecase.training_type != TypeProblem.MultiClassification:
            raise Exception('Confusion matrices only available for multiclassification, not {}'.format(
                            self.usecase.training_type))

        # retrieve current list of predictions if necessary
        if len(self.usecase.predictions) == 0:
            self.usecase.predictions = self.usecase.get_predictions(full=True)

        if predict_id not in self.usecase.predictions:
            raise KeyError('No such prediction')

        prediction = self.usecase.predictions[predict_id]

        # if prediction.truth is None:
        #     raise ValueError('Truth not set for this prediction')

        if not plt:
            raise Exception('matplotlib not installed. please install to plot curves')

        id_col_name = self.usecase.column_config.id_column
        target_col_name = self.usecase.column_config.target_column
        pred_target_col_name = 'pred_' + target_col_name if target_col_name else 'pred_TARGET'
        pred_cols = prediction.drop([id_col_name or 'ID',
                                     target_col_name or 'TARGET',
                                     pred_target_col_name], axis=1)

        pred_labels = pred_cols. \
            idxmax(axis=1). \
            apply(lambda s: s.replace(target_col_name + '_', ''))

        cnf_matrix = confusion_matrix(prediction[target_col_name],
                                      prediction[pred_target_col_name])

        cm = cnf_matrix.astype('float') / cnf_matrix.sum(axis=1)[:, np.newaxis]

        classes = pred_labels.unique()

        plt.imshow(cm, interpolation='nearest', cmap=plt.cm.Blues)
        plt.title('Confusion matrix')
        plt.colorbar()
        tick_marks = np.arange(len(classes))
        plt.xticks(tick_marks, classes, rotation=45)
        plt.yticks(tick_marks, classes)

        fmt = '.2f'
        thresh = cm.max() / 2.
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            plt.text(j, i, format(cm[i, j], fmt),
                     horizontalalignment='center',
                     color='white' if cm[i, j] > thresh else 'black')

        plt.tight_layout()
        plt.show()

    def plot_classif_analysis(self, predict_id, top):
        """
        Plot a classification analysis. Needs a prediction made with the truth present in the dataset.

        Only available for classification.

        Args:
            predict_id (str): ID of the prediction
            top (int): top individuals to analyze

        """
        if self.usecase.training_type != TypeProblem.Classification:
            raise Exception('Classification analysis plots only available for classification, not {}'.format(
                            self.usecase.training_type))

        # retrieve current list of predictions if necessary
        if len(self.usecase.predictions) == 0:
            self.usecase.predictions = self.usecase.get_predictions(full=True)

        if predict_id not in self.usecase.predictions:
            raise KeyError('No such prediction')

        prediction = self.usecase.predictions[predict_id]

        target_col_name = self.usecase.column_config.target_column
        pred_target_col_name = 'pred_' + target_col_name if target_col_name else 'pred_TARGET'

        # sort by predicted values (decreasing)
        prediction.sort_values(pred_target_col_name, ascending=False, inplace=True)
        actual = prediction[target_col_name].values
        predicted = prediction[pred_target_col_name].values

        rec = []
        prec = []
        f1 = []

        # compute indicators
        for i in range(1, top + 1):
            predicted_flag = (predicted >= predicted[i]).astype('int')
            rec.append(recall_score(actual, predicted_flag))
            prec.append(precision_score(actual, predicted_flag))
            f1.append(f1_score(actual, predicted_flag))

        top = np.argmax(f1)

        # plot results
        if not plt:
            raise Exception('matplotlib not installed. please install to plot curves')

        lw = 2

        f, (ax1, ax2, ax3) = plt.subplots(3, 1)

        ax1.plot(rec, 'b-', lw=lw)
        ax1.vlines(x=[top], ymin=0, ymax=1, colors='r', linestyles='solid')
        ax1.set_xlabel('Top individuals')
        ax1.set_ylabel('Recall')
        ax1.set_title('Recall with respect to top n individuals')

        ax2.plot(prec, 'b-', lw=lw)
        ax2.vlines(x=[top], ymin=0, ymax=1, colors='r', linestyles='solid')
        ax2.set_xlabel('Top individuals')
        ax2.set_ylabel('Precision')
        ax2.set_title('Precision with respect to top n individuals')

        ax3.plot(f1, 'b-', lw=lw)
        ax3.vlines(x=[top], ymin=0, ymax=1, colors='r', linestyles='solid')
        ax3.set_xlabel('Top individuals')
        ax3.set_ylabel('F1')
        ax3.set_title('F1 score with respect to top n individuals')

        plt.tight_layout()
        plt.show()

        return {
            'f1': np.round(f1[top], 4),
            'precision': np.round(prec[top], 4),
            'recall': np.round(rec[top], 4),
            'top': (top, 4)
        }
